@@ -1,8 +1,10 @@
 from abc import ABC, abstractmethod
+from collections import UserDict
 from copy import deepcopy
 from functools import reduce
 import math
 from numbers import Number
+import operator
 import random, string
 from sortedcontainers import SortedDict
 from typing import List, Tuple, Optional, Union, Dict, Callable
@@ -341,14 +343,31 @@ class CurveBase(ABC):
         ys = [self[x] for x in xs]
         if kargs.get('label') is None:
             kargs['label']=self.label
-        plt.plot(xs, ys, *args, **kargs)
+        #logger.debug(xs)
+        #logger.debug(ys)
+        line = plt.plot(xs, ys, *args, **kargs)
         kfx = self.keyframes
         kfy = [self[x] for x in kfx]
-        plt.scatter(kfx, kfy)
+        plt.scatter(kfx, kfy, color=line[0].get_color())
 
     def random_label(self):
         return f"curve_{id_generator()}"
 
+    def __sub__(self, other):
+        return self + (-1 * other)
+
+    def __rsub__(self, other):
+        return (-1*self) + other
+
+    def __radd__(self,other) -> 'CurveBase':
+        return self+other
+
+    def __rmul__(self, other) -> 'CurveBase':
+        return self*other
+
+    def __neg__(self):
+        return self * (-1)
+
 
 def id_generator(size=6, chars=string.ascii_uppercase + string.digits):
     # via https://stackoverflow.com/questions/2257441/random-string-generation-with-upper-case-letters-and-digits
@@ -515,6 +534,9 @@ class Curve(CurveBase):
         return {self_label:self, other_label:other}
 
     def __add_curves__(self, other) -> 'Composition':
+        if isinstance(other, ParameterGroup):
+            # this triggers the operator to get resolved by "other" instead of self
+            return NotImplemented
         params = self.__to_labeled(other)
         new_label = '+'.join(params.keys())
         return Composition(parameters=params, label=new_label, reduction='add')
@@ -522,21 +544,22 @@ class Curve(CurveBase):
     def __mul__(self, other) -> CurveBase:
         if isinstance(other, CurveBase):
             return self.__mul_curves__(other)
-        outv = self.copy()
-        for i, k in enumerate(self.keyframes):
-            kf = outv._data[k]
-            kf.value = kf.value * other
-            outv[k]=kf
-        return outv
+        label_ = f"( {other} * {self.label} )"
+        other = Curve(other)
+        return Composition(parameters=(self, other), label=label_, reduction='multiply')
     
     def __mul_curves__(self, other) -> 'Composition':
+        if isinstance(other, ParameterGroup):
+            # this triggers the operator to get resolved by "other" instead of self
+            return NotImplemented
         params = self.__to_labeled(other)
         pg = ParameterGroup(params)
         new_label = '*'.join(params.keys())
         return Composition(parameters=pg, label=new_label, reduction='multiply')
 
-    def __rmul__(self, other) -> 'Curve':
-        return self*other
+    @classmethod
+    def from_function(cls, f:Callable):
+        return cls({0:f(0)}, default_interpolation=lambda k, _: f(k))
 
 
 def SmoothCurve(*args, **kargs):
@@ -548,7 +571,42 @@ def SmoothCurve(*args, **kargs):
     return Curve(*args, default_interpolation='eased_lerp', **kargs)
 
 
-# i'd kind of like this to inherit from dict.
+class DictValuesArithmeticFriendly(UserDict):
+    def __arithmetic_helper(self, operator, other=None):
+        outv = deepcopy(self)
+        for k,v in self.items():
+            if other is not None:
+                outv[k] = operator(v, other)
+            else:
+                outv[k] = operator(v)
+        return outv
+    def __add__(self, other):
+        return self.__arithmetic_helper(operator.add, other)
+    #def __div__(self, other):
+    def __truediv__(self, other): # oh right
+        return self.__arithmetic_helper(operator.truediv, other)
+        #return self.__arithmetic_helper(1/other, operator.mul)
+    def __rtruediv__(self, other):
+        outv = deepcopy(self)
+        for k,v in self.items():
+                outv[k] = other / v
+        return outv
+    def __mul__(self, other):
+        return self.__arithmetic_helper(operator.mul, other)
+    def __neg__(self):
+        return self.__arithmetic_helper(operator.neg)
+    def __radd__(self, other):
+        return self + other
+    def __rmul__(self, other):
+        return self * other
+    def __rsub__(self, other):
+        return (self * (-1)) + other
+    def __sub__(self, other):
+        return self.__arithmetic_helper(operator.sub, other)
+
+
+# i'd kind of like this to inherit from dict. Maybe It can inherit from DictValuesArithmeticFriendly?
+#class ParameterGroup(CurveBase, DictValuesArithmeticFriendly): # Not worth the trouble.
 class ParameterGroup(CurveBase):
     """
     The ParameterGroup class wraps a collection of named parameters to facilitate manipulating them as a unit.
@@ -563,6 +621,7 @@ class ParameterGroup(CurveBase):
     ):
         if isinstance(parameters, list) or isinstance(parameters, tuple):
             d = {}
+            #d = DictValuesArithmeticFriendly()
             for curve in parameters:
                 if not isinstance(curve, CurveBase):
                     curve = Curve(curve)
@@ -578,24 +637,28 @@ class ParameterGroup(CurveBase):
         if not isinstance(weight, Curve):
             weight = Curve(weight)
         self.weight = weight
-        self.parameters={}
+        self.parameters = {}
+        #self.parameters= DictValuesArithmeticFriendly()
         for name, v in parameters.items():
             if not isinstance(v, CurveBase):
                 v = Curve(v)
             v.label = name
             self.parameters[name] = v
         if label is None:
-            label = super().random_label() #self.random_label()
+            label = self.random_label()
         self.label = label
 
     def __getitem__(self, k) -> dict:
         wt = self.weight[k]
-        return {name:param[k]*wt for name, param in self.parameters.items() }
+        d = {name:param[k]*wt for name, param in self.parameters.items() }
+        return DictValuesArithmeticFriendly(d)
+        #return (wt * self.parameters)[k] #
 
     # this might cause performance issues down the line. deal with it later.
     def copy(self) -> 'ParameterGroup':
         return deepcopy(self)
 
+    # feels a bit redundant with DictValuesArithmeticFriendly, but fuck it.
     def __add__(self, other) -> 'ParameterGroup':
         outv = self.copy()
         for k,v in outv.parameters.items():
@@ -607,21 +670,32 @@ class ParameterGroup(CurveBase):
         for k,v in outv.parameters.items():
             outv.parameters[k] = v * other
         return outv
-
-    def __radd__(self,other) -> 'ParameterGroup':
-        return self+other
-
-    def __rmul__(self, other) -> 'ParameterGroup':
-        return self*other
+    
+    def __truediv__(self, other) -> 'ParameterGroup':
+        logger.debug(self.label)
+        logger.debug(other.label)
+        outv = self.copy()
+        for k,v in outv.parameters.items():
+            outv.parameters[k] = v / other
+        return outv
+    
+    def __rtruediv__(self, other) -> 'ParameterGroup':
+        outv = self.copy()
+        for k,v in outv.parameters.items():
+            outv.parameters[k] = other / v
+        return outv
 
     @property
     def duration(self):
         return max(curve.duration for curve in self.parameters.values())
 
-    def plot(self):
-        n = self.duration + 1
+    def plot(self, n:int=None, xs:list=None, eps:float=1e-9, *args, **kargs):
+        if n is None:
+            n = self.duration + 1
         for curve in self.parameters.values():
-            curve.plot(n=n)
+            curve = curve.copy()
+            curve = curve * self.weight
+            curve.plot(n=n, xs=xs, eps=eps, *args, **kargs)
 
     @property
     def keyframes(self):
@@ -636,21 +710,24 @@ class ParameterGroup(CurveBase):
     def values(self):
         return [self[k] for k in self.keyframes]
 
-import operator
+    def random_label(self):
+        return f"pgroup({','.join([c.label for c in self.parameters.values()])})"
+
 
 REDUCTIONS = {
-    'add': operator.add, #lambda x,y:x+y,
-    'sum': operator.add, # lambda x,y:x+y,
-    'multiply': operator.mul, # lambda x,y:x*y,
-    'product': operator.mul, # lambda x,y:x*y,
-    'prod': operator.mul, # lambda x,y:x*y, # what is wrong with me...
-    'subtract': operator.sub, # lambda x,y:x-y,
-    'sub': operator.sub, # lambda x,y:x-y,
-    'divide': operator.truediv, # lambda x,y:x/y,
-    'div': operator.truediv, # lambda x,y:x/y,
+    'add': operator.add,
+    'sum': operator.add,
+    'multiply': operator.mul,
+    'product': operator.mul,
+    'prod': operator.mul,
+    'subtract': operator.sub,
+    'sub': operator.sub,
+    'divide': operator.truediv,
+    'div': operator.truediv,
+    'truediv': operator.truediv,
     'max':max,
     'min':min,
-    ## require special treatment by caller
+    ## requires special treatment by caller
     'mean': operator.add,
     'average': operator.add,
     'avg': operator.add,
@@ -677,13 +754,18 @@ class Composition(ParameterGroup):
         reduction:str=None,
         label:str=None,
     ):
-        super().__init__(parameters=parameters, weight=weight, label=label)
         self.reduction = reduction
+        _label = label
+        super().__init__(parameters=parameters, weight=weight, label=label)
+        if _label is None:
+            self.label = self.random_label()
 
     def __getitem__(self, k):
         f = REDUCTIONS.get(self.reduction)
 
         vals = [curve[k] for curve in self.parameters.values()]
+        #logger.debug(self.label)
+        #logger.debug(vals)
         outv = reduce(f, vals)
         if self.reduction in ('avg', 'average', 'mean'):
             outv = outv * (1/ len(vals))
@@ -691,14 +773,12 @@ class Composition(ParameterGroup):
         return outv
 
     def random_label(self):
-        basename = ' '.join(self.parameters.keys())
+        basename = ', '.join(self.parameters.keys())
         return f"{self.reduction}({basename})_{id_generator()}"
 
     def __radd__(self, other):
         return super().__radd__(other)
 
-
-
     def __add__(self, other) -> 'Composition':
         if not isinstance(other, CurveBase):
             other = Curve(other)
@@ -713,8 +793,6 @@ class Composition(ParameterGroup):
             d = {pg_copy.label:pg_copy, other.label:other}
             return Composition(parameters=d, weight=pg_copy.weight, reduction='sum')
 
-
-
     def __mul__(self, other) -> 'ParameterGroup':
         if not isinstance(other, CurveBase):
             other = Curve(other)
@@ -729,4 +807,70 @@ class Composition(ParameterGroup):
             return pg_copy
         else:
             d = {pg_copy.label:pg_copy, other.label:other}
-            return Composition(parameters=d, reduction='prod')
\ No newline at end of file
+            return Composition(parameters=d, reduction='prod')
+
+    # to do: abstract a help for consistency...
+    def __truediv__(self, other) -> 'ParameterGroup':
+        if not isinstance(other, CurveBase):
+            other = Curve(other)
+            if (other.label in self.parameters) or (other.label == self.label):
+                other.label = other.random_label()
+        pg_copy = self.copy()
+        d = {pg_copy.label:pg_copy, other.label:other}
+        return Composition(parameters=d, reduction='truediv')
+    
+    def __rtruediv__(self, other):
+        if not isinstance(other, CurveBase):
+            other = Curve(other)
+            if (other.label in self.parameters) or (other.label == self.label):
+                other.label = other.random_label()
+        pg_copy = self.copy()
+        #d = {pg_copy.label:pg_copy, other.label:other}
+        #return Composition(parameters=d, reduction='truediv')
+        d = {other.label:other, pg_copy.label:pg_copy} # reverse order of arguments
+        return Composition(parameters=d, reduction='truediv')
+
+    def plot(self, n:int=None, xs:list=None, eps:float=1e-9, *args, **kargs):
+        """
+        Arguments
+            n (int): (Optional) Number of points to plot, plots range [0,n-1]. If not specified, n=self.duration.
+            eps (float): (Optional) value to be subtracted from keyframe to produce additional points for plotting.
+                Plotting these additional values is important for e.g. visualizing step function behavior.
+        """
+        is_compositional_pgroup = False
+        for v in self.parameters.values():
+            if isinstance(v, ParameterGroup) and not isinstance(v, Composition):
+                is_compositional_pgroup = True
+                break
+        if not is_compositional_pgroup:
+            Curve.plot(self, n=n, xs=xs, eps=eps, *args, **kargs)
+        else: 
+            #raise NotImplementedError
+            try:
+                import matplotlib.pyplot as plt
+            except ImportError:
+                raise ImportError("Please install matplotlib to use Curve.plot()")
+
+            if xs is None:
+                if n is None:
+                    n = self.duration + 1
+                xs_base = list(range(int(n))) + list(self.keyframes)
+                xs = set()
+                for x in xs_base:
+                    if (x>0) and (eps is not None) and (eps > 0):
+                        xs.add(x-eps)
+                    xs.add(x)
+            xs = list(set(xs))
+            xs.sort()
+            
+            ys = [self[x] for x in xs]
+            # https://stackoverflow.com/questions/5558418/list-of-dicts-to-from-dict-of-lists
+            ys_d =  {k: [dic[k] for dic in ys] for k in ys[0]}
+            for label, values in ys_d.items():
+                #if kargs.get('label') is None:
+                #    kargs['label']=self.label
+                kargs['label'] = label
+                plt.plot(xs, values, *args, **kargs)
+                kfx = self.keyframes
+                kfy = [self[x][label] for x in kfx]
+                plt.scatter(kfx, kfy)
